''' GSuite '''
from __future__ import print_function

import re
from typing import Any, Generator, Union

from google.oauth2 import service_account  # type: ignore
from googleapiclient import errors  # type: ignore
from googleapiclient.discovery import build  # type: ignore

RE_PICTURE = re.compile(r'(https://.+){1,}=?s([\d]{1,}-c)')


class GSuite:
    ''' GSuite '''
    __slots__ = ('service', )

    SCOPES = ('https://www.googleapis.com/auth/admin.directory.user',
              'https://www.googleapis.com/auth/admin.directory.group',
              )

    def __init__(self, credentialfile: str, with_subject: str):
        creds = service_account.Credentials.from_service_account_file(
            credentialfile, scopes=self.SCOPES).with_subject(with_subject)

        self.service = build('admin', 'directory_v1',
                             credentials=creds, cache_discovery=False)

    @property
    def print_scopes(self) -> str:
        ''' Print the scopes '''
        return ','.join(self.SCOPES)

    # ----- Users.list ----- #
    # https://googleapis.github.io/google-api-python-client/docs/dyn/admin_directory_v1.users.html#list
    def users_list(self) -> Any:
        ''' Users.list '''
        return self.service.users().list(customer='my_customer', orderBy='email').execute()

    # ----- Users.get ----- #
    # https://googleapis.github.io/google-api-python-client/docs/dyn/admin_directory_v1.users.html#get
    def users_get(self, user_key: str) -> Any:
        ''' Users.get '''
        return self.service.users().get(userKey=user_key).execute()

    # ----- Groups ----- #
    # https://developers.google.com/admin-sdk/directory/v1/reference/groups
    # https://googleapis.github.io/google-api-python-client/docs/dyn/admin_directory_v1.groups.html
    def groups_get(self, group_key: str) -> Any:
        ''' Groups.get '''
        return self.service.groups().get(groupKey=group_key).execute()

    # ----- Groups.list ----- #
    # https://googleapis.github.io/google-api-python-client/docs/dyn/admin_directory_v1.groups.html#list
    def groups_list(self, page_token: Union[str, None] = None) -> Any:
        ''' Groups.list '''
        return self.service.groups().list(customer='my_customer',
                                          orderBy='email', pageToken=page_token).execute()

    def groups_list_loop(self, page_token: Union[str, None] = None) ->\
            Generator[dict[str, str], None, None]:
        ''' Groups.list.loop '''
        groups = self.groups_list(page_token=page_token)
        for group in groups['groups']:
            yield group

        if 'nextPageToken' in groups:
            for group in self.groups_list_loop(page_token=groups['nextPageToken']):
                yield group

    def groups_insert(self, email: str, description: Union[str, None] = None,
                      name: Union[str, None] = None) -> Any:
        ''' Groups.insert '''
        body = {'email': email}
        if description is not None:
            body['description'] = description

        if name is not None:
            body['name'] = name

        return self.service.groups().insert(body=body).execute()

    # ----- Members ----- #
    # https://developers.google.com/admin-sdk/directory/v1/reference/members
    # https://googleapis.github.io/google-api-python-client/docs/dyn/admin_directory_v1.members.html
    def members_list(self, group_key: str, page_token: Union[str, None] = None) -> Any:
        ''' members.list '''
        return self.service.members().list(groupKey=group_key, pageToken=page_token).execute()

    def members_list_loop(self, group_key: str) -> Generator[dict[str, str], None, None]:
        ''' members.list.loop '''
        members = self.members_list(group_key)
        for member in members.get('members', []):
            yield member

        while 'nextPageToken' in members:
            members = self.members_list(
                group_key, page_token=members['nextPageToken'])
            for member in members.get('members', []):
                yield member

    def members_insert(self, group_key: str, email: str,
                       role: str = 'MEMBER', delivery_settings: str = 'ALL_MAIL') -> Any:
        ''' members.insert '''
        body = {'email': email, 'role': role,
                'delivery_settings': delivery_settings}
        return self.service.members().insert(groupKey=group_key, body=body).execute()

    def members_has_member(self, group_key: str, email: str) -> dict[str, bool]:
        ''' members.hasMember '''
        try:
            if self.members_get(group_key=group_key, email=email):
                return {'isMember': True}
            return {'isMember': False}
        except errors.HttpError:
            return {'isMember': False}

    def members_get(self, group_key: str, email: str) -> Any:
        ''' members.get '''
        return self.service.members().get(groupKey=group_key, memberKey=email).execute()

    def members_delete(self, group_key: str, email: str) -> Any:
        ''' members.delete '''
        return self.service.members().delete(groupKey=group_key, memberKey=email).execute()

    @staticmethod
    def size_picture(url: str, size: int = 512) -> str:
        ''' Convert picture size '''
        result = RE_PICTURE.match(url)

        if result:
            _url, _size = result.groups()
            return url.replace(f's{_size}', f's{size}-c')

        return url



''' AWS SDK API Integration

:Integration:
    - S3 :class:`AWSS3`
    - SES :class:`AWSSES`

'''
from email import encoders
from email.charset import Charset
from email.mime.base import MIMEBase
from email.mime.multipart import MIMEMultipart
from email.mime.text import MIMEText
from email.utils import formataddr
from os.path import basename
from typing import Any

import boto3  # type: ignore


class AWSS3:
    ''' AWSS3

    :param str aws_access_key_id: aws_access_key_id
    :param str aws_secret_access_key: aws_secret_access_key
    :param str bucket: bucket name

    '''
    __slots__ = ('client', 'bucket')

    def __init__(self, aws_access_key_id: str, aws_secret_access_key: str, bucket: str) -> None:
        self.client = boto3.client('s3',
                                   aws_access_key_id=aws_access_key_id,
                                   aws_secret_access_key=aws_secret_access_key,
                                   )
        self.bucket = bucket

    def get_object(self, key: str) -> Any:
        ''' Get object

        :param str key: key name
        :rtype: dict

        .. seealso::
            :meth:`S3.Client.get_object`

        '''
        return self.client.get_object(Bucket=self.bucket, Key=key)

    def convert_to_attachment(self, key: str) -> MIMEBase:
        ''' Convert the object to match attachment '''
        s3object = self.get_object(key)
        attachment = MIMEBase(
            s3object['ContentType'].split('/')[0],
            f'''{s3object['ContentType'].split('/')[1]}; name="{Charset('utf-8').header_encode(basename(key))}"'''  # pylint: disable=line-too-long
        )
        attachment.add_header(
            'Content-Disposition',
            f'''attachment; filename="{Charset('utf-8').header_encode(basename(key))}"''')
        attachment.set_payload(s3object['Body'].read())
        encoders.encode_base64(attachment)

        return attachment


class AWSSES:
    ''' AWSSES

    :param str aws_access_key_id: aws_access_key_id
    :param str aws_secret_access_key: aws_secret_access_key
    :param dict source: ``{'name': '', 'mail': ''}``, mail's ``from``

    '''

    __slots__ = ('client', 'source')

    def __init__(self, aws_access_key_id: str, aws_secret_access_key: str,
                 source: dict[str, str]) -> None:
        self.client = boto3.client('ses',
                                   aws_access_key_id=aws_access_key_id,
                                   aws_secret_access_key=aws_secret_access_key,
                                   region_name='us-east-1')
        self.source = source

    @staticmethod
    def format_mail(name: str, mail: str) -> str:
        ''' Encode header to base64

            :param str name: user name
            :param str mail: user mail
            :rtype: str
            :return: a string of ``name <mail>`` in ``base64``.
        '''
        if name:
            return formataddr((name, mail))

        return mail

    def send_email(self, *args: Any, **kwargs: Any) -> Any:
        ''' Send mail

        ``*args``, ``**kwargs`` are the same with :meth:`SES.Client.send_email`

        .. seealso::
            :meth:`SES.Client.send_email`

        '''
        return self.client.send_email(*args, **kwargs)

    def raw_mail(self, **kwargs: Any) -> Any:
        ''' To make raw mail content

        :param list to_addresses: to
        :param str subject: subject
        :param str body: body
        :param str x_coscup: (optional) for mail header ``X-Coscup``
        :param list cc_addresses: (optional) cc
        :param list attachment: (optional) attachment

        :rtype: :py:class:`email.mime.multipart.MIMEMultipart`

        '''
        msg_all = MIMEMultipart()

        msg_all['From'] = self.format_mail(
            name=self.source['name'], mail=self.source['mail'])

        to_list = []
        for to_user in kwargs['to_addresses']:
            to_list.append(self.format_mail(to_user['name'], to_user['mail']))
        msg_all['To'] = ','.join(to_list)

        cc_list = []
        if 'cc_addresses' in kwargs and kwargs['cc_addresses']:
            for cc_user in kwargs['cc_addresses']:
                cc_list.append(self.format_mail(
                    cc_user['name'], cc_user['mail']))

            if cc_list:
                msg_all['Cc'] = ','.join(cc_list)

        msg_all['Subject'] = kwargs['subject']

        if 'x_coscup' in kwargs and kwargs['x_coscup']:
            msg_all['X-Coscup'] = kwargs['x_coscup']

        msg_all.attach(MIMEText(kwargs['body'], 'html', 'utf-8'))

        if 'attachment' in kwargs and kwargs['attachment']:
            for attach in kwargs['attachment']:
                msg_all.attach(attach)

        return msg_all

    def send_raw_email(self, **kwargs: Any) -> Any:
        ''' send raw email

        if no ``data``, must have ``from``, ``to``, ``subject``, ``body``, ``attachment``

        :param str source: (optional) from
        :param str to_addresses: (optional) to
        :param str subject: (optional) subject
        :param str body: (optional) body
        :param list attachment: (optional) attachment
        :param data: (optional) MIMEMultipart data
        :param data_str: (optional) MIMEMultipart to str data
        :type data: :py:class:`email.mime.multipart.MIMEMultipart`

        .. seealso::
            :meth:`SES.Client.send_raw_email`

        '''
        if 'data_str' in kwargs:
            return self.client.send_raw_email(
                RawMessage={'Data': kwargs['data_str']})

        data = kwargs.get('data')
        if not data:
            data = self.raw_mail(**kwargs)

        return self.client.send_raw_email(
            RawMessage={'Data': data.as_string()})



''' Mattermost '''
# pylint: disable=arguments-renamed,arguments-differ
import logging
from typing import Any, Generator, Optional

from requests import Response, Session


class Mattermost(Session):
    ''' Mattermost '''

    def __init__(self, token: str, base_url: str, log_name: str = 'Mattermost'):
        super().__init__()
        self.token = token
        self.base_url = base_url
        self.log = logging.getLogger(log_name)
        self.headers.update({'Authorization': f'Bearer {self.token}'})

    def log_rate_limit(self, headers: dict[str, Any]) -> None:
        ''' Get log info from headers '''
        self.log.info('X-Ratelimit-Limit: %s, X-Ratelimit-Remaining: %s, X-Ratelimit-Reset: %s',
                      headers.get('X-Ratelimit-Limit'),
                      headers.get('X-Ratelimit-Remaining'),
                      headers.get('X-Ratelimit-Reset'),
                      )

    def get_users(self, page: int, per_page: int = 200) -> Response:
        ''' Get users '''
        return self.get(f'{self.base_url}/users', params={'page': page, 'per_page': per_page})

    def get_users_loop(self, per_page: int = 200) -> Generator[dict[str, Any], None, None]:
        ''' Get users in loop '''
        page = 0
        num = 0
        for user in self.get_users(page=page, per_page=per_page).json():
            yield user
            num += 1

        while num == per_page:
            page += 1
            num = 0
            for user in self.get_users(page=page, per_page=per_page).json():
                yield user
                num += 1

    def get_users_stats(self) -> Response:
        ''' Get users stats '''
        return self.get(f'{self.base_url}/users/stats')

    def get_user_by_username(self, username: str) -> Response:
        ''' Get user by username '''
        return self.get(f'{self.base_url}/users/username/{username}')

    def create_a_direct_message(self, users: str) -> Response:
        ''' Create a direct messge '''
        return self.post(f'{self.base_url}/channels/direct', json=users)

    def posts(self, channel_id: str, message: str) -> Response:
        ''' Posts message '''
        return self.post(f'{self.base_url}/posts',
                         json={'channel_id': channel_id, 'message': message})

    def get_posts_from_channel(self, channel_id: str) -> Response:
        ''' Get post from channel '''
        return self.get(f'{self.base_url}/channels/{channel_id}/posts')

    def post_invite_by_email(self, team_id: str, emails: list[str]) -> Response:
        ''' Post an invite by email '''
        return self.post(f'{self.base_url}/teams/{team_id}/invite/email', json=emails)

    def post_invite_guests_by_email(self, team_id: str, emails: list[str],
                                    channels: list[str], message: Optional[str] = None) -> Response:
        ''' Post an invite to guest by email '''
        data: dict[str, Any] = {
            'emails': emails,
            'channels': channels,
        }
        if message:
            data['message'] = message.strip()

        return self.post(f'{self.base_url}/teams/{team_id}/invite-guests/email', json=data)

    def post_user_to_channel(self, channel_id: str, uid: str) -> Response:
        ''' Post user to channel '''
        return self.post(f'{self.base_url}/channels/{channel_id}/members', json={'user_id': uid})

    def put_users_patch(self, uid: str, position: str) -> Response:
        ''' Update user '''
        data = {
            'position': position,
        }
        return self.put(f'{self.base_url}/users/{uid}/patch', json=data)



''' IPInfo '''
from requests import Response, Session


class IPInfo(Session):
    ''' IPInfo '''

    def __init__(self, token: str):
        super().__init__()
        self.url = 'https://ipinfo.io/'
        self.token = token
        self.headers.update({'Authorization': f'Bearer {self.token}'})

    def get_info(self, ip_address: str) -> Response:
        ''' Get info

        :param str ip: ip

        '''
        return super().get(f'{self.url}{ip_address}')



''' TelegramBot '''
from typing import Any, Union

from requests import Response, Session


class Telegram(Session):
    ''' Telegram '''

    def __init__(self, token: str):
        super().__init__()
        self.url = f'https://api.telegram.org/bot{token}'

    def get_me(self) -> Response:
        ''' Get me '''
        return self.post(f'{self.url}/getMe')

    def send_message(self, chat_id: str, text: str,
                     parse_mode: str = 'Markdown',
                     reply_markup: Union[dict[str, Any], None] = None) -> Response:
        ''' Send message '''
        data = {
            'chat_id': chat_id,
            'text': text,
            'parse_mode': parse_mode}  # type: dict[str, Union[str, dict[str, Any]]]

        if reply_markup is not None:
            data['reply_markup'] = reply_markup

        return self.post(f'{self.url}/sendMessage', json=data)

    def set_webhook(self, url: str) -> Response:
        ''' Set webhook '''
        return self.post(f'{self.url}/setWebhook', json={'url': url})

    def get_webhook_info(self, url: str) -> Response:
        ''' Get webhook info '''
        return self.post(f'{self.url}/getWebhookInfo', json={'url': url})

    def delete_webhook(self) -> Response:
        ''' delete webhook '''
        return self.post(f'{self.url}/deleteWebhook')

    @staticmethod
    def is_command_start(data: dict[str, Any]) -> bool:
        ''' command start '''
        if data['message']['from']['is_bot']:
            return False

        if data['message']['text'].strip() == '/start':
            return True

        return False

    @staticmethod
    def is_command_start_linkme(data: dict[str, Any]) -> bool:
        ''' command start '''
        if 'message' not in data:
            return False

        if data['message']['from']['is_bot']:
            return False

        if data['message']['text'].strip() in ('/start linkme', '/linkme'):
            return True

        return False



''' GitlabAPI '''
from requests import Response, Session


class GitlabAPI(Session):
    ''' GitlabAPI '''

    def __init__(self, token: str) -> None:
        super().__init__()
        self.url = 'https://gitlab.com/api/v4'
        self.token = token
        self.headers.update({'Authorization': f'Bearer {self.token}'})

    def get_project(self, project_id: str) -> Response:
        ''' Get Project '''
        return self.get(url=f'{self.url}/projects/{project_id}')

    def post_invite_to_project(self, project_id: str, email: str,
                               access_level: int = 30) -> Response:
        ''' Post invite to project

        access_level 30 => Developer

        https://docs.gitlab.com/ee/api/invitations.html#valid-access-levels
        https://docs.gitlab.com/ee/user/permissions.html#project-members-permissions
        '''
        return self.post(url=f'{self.url}/projects/{project_id}/invitations',
                         data={'email': email, 'access_level': access_level})

    def delete_invite_to_project(self, project_id: str, email: str) -> Response:
        ''' DELETE invite to project '''
        return self.delete(url=f'{self.url}/projects/{project_id}/invitations/{email}')
